// Software-PWM by PoWl - 25.10.08
// webmaster@the-powl.de

// ATmega8 @ 8Mhz

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdbool.h>


// Sinustabelle fr Demo
const uint8_t sinetable[] = {
128, 131, 134, 137, 140, 143, 146, 149, 152, 155, 158, 162, 165, 167,
170, 173, 176, 179, 182, 185, 188, 190, 193, 196, 198, 201, 203, 206,
208, 211, 213, 215, 218, 220, 222, 224, 226, 228, 230, 232, 234, 235,
237, 238, 240, 241, 243, 244, 245, 246, 248, 249, 250, 250, 251, 252,
253, 253, 254, 254, 254, 255, 255, 255, 255, 255, 255, 255, 254, 254,
254, 253, 253, 252, 251, 250, 250, 249, 248, 246, 245, 244, 243, 241,
240, 238, 237, 235, 234, 232, 230, 228, 226, 224, 222, 220, 218, 215,
213, 211, 208, 206, 203, 201, 198, 196, 193, 190, 188, 185, 182, 179,
176, 173, 170, 167, 165, 162, 158, 155, 152, 149, 146, 143, 140, 137,
134, 131, 128, 124, 121, 118, 115, 112, 109, 106, 103, 100, 97, 93,
90, 88, 85, 82, 79, 76, 73, 70, 67, 65, 62, 59, 57, 54, 52, 49, 47,
44, 42, 40, 37, 35, 33, 31, 29, 27, 25, 23, 21, 20, 18, 17, 15, 14,
12, 11, 10, 9, 7, 6, 5, 5, 4, 3, 2, 2, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0,
1, 1, 1, 2, 2, 3, 4, 5, 5, 6, 7, 9, 10, 11, 12, 14, 15, 17, 18, 20,
21, 23, 25, 27, 29, 31, 33, 35, 37, 40, 42, 44, 47, 49, 52, 54, 57,
59, 62, 65, 67, 70, 73, 76, 79, 82, 85, 88, 90, 93, 97, 100, 103,
106, 109, 112, 115, 118, 121, 124 };



// -------------------------- 8-Bit PWM --------------------------- //
// An - 254 Graustufen - Aus

#define CHANNELS 8
#define PWM_PORT PORTB

uint8_t pwm_pin[CHANNELS] = { PB0, PB1, PB2, PB3, PB4, PB5, PB6, PB7 };

uint8_t pwm_mask_A[CHANNELS + 1];
uint8_t pwm_timing_A[CHANNELS + 1];

uint8_t pwm_mask_B[CHANNELS + 1];
uint8_t pwm_timing_B[CHANNELS + 1];

uint8_t *ptr_PORT_main = &pwm_mask_A;
uint8_t *ptr_timing_main = &pwm_timing_A;

uint8_t *ptr_PORT_isr = &pwm_mask_B;
uint8_t *ptr_timing_isr = &pwm_timing_B;

uint8_t pwm_value[CHANNELS];
uint8_t pwm_queue[CHANNELS];

volatile uint8_t pwm_cycle;
volatile bool pwm_change;


void pwm_update(void)
{
  pwm_change = 0;

  // Sortieren mit InsertSort

  for(uint8_t i=1; i<CHANNELS; i++)
  {
    uint8_t j = i;

    while(j && pwm_value[pwm_queue[j - 1]] > pwm_value[pwm_queue[j]])
    {
      uint8_t temp = pwm_queue[j - 1];
      pwm_queue[j - 1] = pwm_queue[j];
      pwm_queue[j] = temp;

      j--;
    }
  }

  // Maske generieren

  for(uint8_t i=0; i<(CHANNELS + 1); i++)
  {
    ptr_PORT_main[i] = 0;
    ptr_timing_main[i] = 0;
  }

  uint8_t j = 0;

  for(uint8_t i=0; i<CHANNELS; i++)
  {
    // Sonderfall "ganz AUS" ausschlieen:
    //  - Pins garnicht erst anschalten
    //  - Pins die vorher "ganz AN" waren mssen trotzdem noch deaktiviert werden, deshalb nur folgendes in der if
    if(pwm_value[pwm_queue[i]] != 0)
    {
      ptr_PORT_main[0] |= (1 << pwm_pin[pwm_queue[i]]);
    }

    // Sonderfall "ganz AN" ausschlieen:
    //  - Nicht in Timing-Tabelle bernehmen da sonst die Pointer niemals getauscht werden
    //  - Pins die "ganz AN" sind mssen nicht abgeschaltet werden
    if(pwm_value[pwm_queue[i]] != 255)
    {
      ptr_PORT_main[j + 1] |= (1 << pwm_pin[pwm_queue[i]]);

      ptr_timing_main[j] = pwm_value[pwm_queue[i]];
    }

    if(pwm_value[pwm_queue[i]] != pwm_value[pwm_queue[i + 1]])
    {
      j++;
    }
  }

  /*
  // Der ISR etwas Arbeit abnehmen
  for(uint8_t i=1; i<(CHANNELS + 1); i++)
  {
    ptr_PORT_main[i] = ~ptr_PORT_main[i];
  }
  */

  pwm_change = 1;
}


int main(void)
{
  // PWM initialisieren
  for(uint8_t i=0; i<CHANNELS; i++)
  {
    pwm_queue[i] = i;
  }

  DDRB = (1 << PB0) | (1 << PB1) | (1 << PB2) | (1 << PB3) | (1 << PB4) | (1 << PB5) | (1 << PB6) | (1 << PB7);


  // Timer fr PWM
  OCR1A = 255;
  TIMSK = (1 << OCIE1A) | (1 << OCIE1B);
  TCCR1B = (1 << WGM12) | 4;

  sei();


  uint8_t pos = 0;

  while(1)
  {
    for(uint8_t i=0; i<8; i++)
    {
      pwm_value[i] = sinetable[(uint8_t) (pos + i * 20)];
    }

    pwm_update();
    pos++;

    _delay_ms(5);
  }
}



ISR(TIMER1_COMPA_vect)
{
  pwm_cycle = 0;

  PWM_PORT |= ptr_PORT_isr[0];
  OCR1B = ptr_timing_isr[0];
}



ISR(TIMER1_COMPB_vect)
{
  pwm_cycle++;

  PWM_PORT &= ~ptr_PORT_isr[pwm_cycle];

  if(ptr_timing_isr[pwm_cycle] != 0)
  {
    OCR1B = ptr_timing_isr[pwm_cycle];
  }
  else if(pwm_change)
  {
    pwm_change = 0;

    uint8_t *temp_ptr;

    temp_ptr = ptr_PORT_main;
    ptr_PORT_main = ptr_PORT_isr;
    ptr_PORT_isr = temp_ptr;

    temp_ptr = ptr_timing_main;
    ptr_timing_main = ptr_timing_isr;
    ptr_timing_isr = temp_ptr;
  }
}
